"""Render descriptions to C# documentation comments."""
import textwrap
from typing import Tuple, Optional, List
import xml.sax.saxutils

from icontract import ensure
import docutils.nodes
import docutils.parsers.rst.roles
import docutils.utils

from aas_core_codegen.common import Stripped, Error, assert_never, Identifier
from aas_core_codegen import intermediate
from aas_core_codegen.intermediate import (
    doc as intermediate_doc,
    rendering as intermediate_rendering,
)
from aas_core_codegen.csharp import (
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import INDENT as I


class _ElementRenderer(intermediate_rendering.DocutilsElementTransformer[str]):
    """Render descriptions as C# docstring XML."""

    def transform_text(
        self, element: docutils.nodes.Text
    ) -> Tuple[Optional[str], Optional[str]]:
        return xml.sax.saxutils.escape(element.astext()), None

    def transform_symbol_reference_in_doc(
        self, element: intermediate_doc.SymbolReference
    ) -> Tuple[Optional[str], Optional[str]]:
        name = None  # type: Optional[str]

        if isinstance(element.symbol, intermediate.Enumeration):
            name = csharp_naming.enum_name(element.symbol.name)

        elif isinstance(element.symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2021-12-17):
            # We do not generate a class for constrained primitives, but we
            # leave it as class name, as that is what we used for ``Verify*`` function.
            name = csharp_naming.class_name(element.symbol.name)

        elif isinstance(element.symbol, intermediate.Class):
            if isinstance(element.symbol, intermediate.AbstractClass):
                # NOTE (mristin, 2021-12-25):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                name = csharp_naming.interface_name(element.symbol.name)

            elif isinstance(element.symbol, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.
                name = csharp_naming.class_name(element.symbol.name)

            else:
                assert_never(element.symbol)

        else:
            assert_never(element.symbol)

        assert name is not None
        return f"<see cref={xml.sax.saxutils.quoteattr(name)} />", None

    def transform_attribute_reference_in_doc(
        self, element: intermediate_doc.AttributeReference
    ) -> Tuple[Optional[str], Optional[str]]:
        cref = None  # type: Optional[str]

        if isinstance(element.reference, intermediate_doc.PropertyReference):
            symbol_name = None  # type: Optional[str]

            if isinstance(element.reference.cls, intermediate.AbstractClass):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                symbol_name = csharp_naming.interface_name(element.reference.cls.name)
            elif isinstance(element.reference.cls, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.

                symbol_name = csharp_naming.class_name(element.reference.cls.name)
            else:
                assert_never(element.reference.cls)

            prop_name = csharp_naming.property_name(element.reference.prop.name)

            assert symbol_name is not None
            cref = f"{symbol_name}.{prop_name}"
        elif isinstance(
            element.reference, intermediate_doc.EnumerationLiteralReference
        ):
            symbol_name = csharp_naming.enum_name(element.reference.symbol.name)
            literal_name = csharp_naming.enum_literal_name(
                element.reference.literal.name
            )

            cref = f"{symbol_name}.{literal_name}"
        else:
            assert_never(element.reference)

        assert cref is not None
        return f"<see cref={xml.sax.saxutils.quoteattr(cref)} />", None

    def transform_argument_reference_in_doc(
        self, element: intermediate_doc.ArgumentReference
    ) -> Tuple[Optional[str], Optional[str]]:
        arg_name = csharp_naming.argument_name(Identifier(element.reference))
        return f"<paramref name={xml.sax.saxutils.quoteattr(arg_name)} />", None

    def transform_literal(
        self, element: docutils.nodes.literal
    ) -> Tuple[Optional[str], Optional[str]]:
        return f"<c>{xml.sax.saxutils.escape(element.astext())}</c>", None

    def transform_paragraph(
        self, element: docutils.nodes.paragraph
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "".join(parts), None

    def transform_emphasis(
        self, element: docutils.nodes.emphasis
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "<em>{}</em>".format("".join(parts)), None

    def transform_list_item(
        self, element: docutils.nodes.list_item
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "<li>{}</li>".format("".join(parts)), None

    def transform_bullet_list(
        self, element: docutils.nodes.bullet_list
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = ["<ul>\n"]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(f"{text}\n")
        parts.append("</ul>")

        return "".join(parts), None

    def transform_note(
        self, element: docutils.nodes.note
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "".join(parts), None

    def transform_reference(
        self, element: docutils.nodes.reference
    ) -> Tuple[Optional[str], Optional[str]]:
        parts = []  # type: List[str]
        for child in element.children:
            text, error = self.transform(child)
            if error is not None:
                return None, error

            assert text is not None
            parts.append(text)

        return "".join(parts), None

    def transform_document(
        self, element: docutils.nodes.document
    ) -> Tuple[Optional[str], Optional[str]]:
        if len(element.children) == 0:
            return "", None

        summary = None  # type: Optional[docutils.nodes.paragraph]
        remarks = []  # type: List[docutils.nodes.Element]
        tail = []  # type: List[docutils.nodes.Element]

        # Try to match the summary and the remarks
        if len(element.children) >= 1:
            if not isinstance(element.children[0], docutils.nodes.paragraph):
                return None, (
                    f"Expected the first document element to be a summary and "
                    f"thus a paragraph, but got: {element.children[0]}"
                )

            summary = element.children[0]

        remainder = element.children[1:]
        for i, child in enumerate(remainder):
            if isinstance(
                child,
                (
                    docutils.nodes.paragraph,
                    docutils.nodes.bullet_list,
                    docutils.nodes.note,
                ),
            ):
                remarks.append(child)
            else:
                tail = remainder[i:]
                break

        # NOTE (2021-09-16, mristin):
        # We restrict ourselves here quite a lot. This function will need to evolve as
        # we add a larger variety of docstrings to the meta-model.
        #
        # For example, we need to translate ``:paramref:``'s to ``<paramref ...>`` in
        # C#. Additionally, we need to change the name of the argument accordingly
        # (``snake_case`` to ``camelCase``).

        # Blocks to be joined by a new-line
        blocks = []  # type: List[Stripped]

        renderer = _ElementRenderer()

        if summary:
            summary_text, error = renderer.transform(element=summary)
            if error:
                return None, error

            assert summary_text is not None
            blocks.append(Stripped(f"<summary>\n" f"{summary_text}\n" f"</summary>"))

        if remarks:
            remark_blocks = []  # type: List[str]
            for remark in remarks:
                remark_text, error = renderer.transform(element=remark)
                if error:
                    return None, error

                assert remark_text is not None
                remark_blocks.append(remark_text)

            assert len(remark_blocks) >= 1, (
                f"Expected at least one remark block "
                f"since ``remarks`` defined: {remarks}"
            )

            if len(remark_blocks) == 1:
                blocks.append(
                    Stripped(f"<remarks>\n" f"{remark_blocks[0]}\n" f"</remarks>")
                )
            else:
                remarks_paras = "\n".join(
                    f"<para>{remark_block}</para>" for remark_block in remark_blocks
                )

                blocks.append(
                    Stripped(f"<remarks>\n" f"{remarks_paras}\n" f"</remarks>")
                )

        for tail_element in tail:
            # BEFORE-RELEASE (mristin, 2021-12-13): test
            if not isinstance(tail_element, docutils.nodes.field_list):
                return None, (
                    f"Expected only a field list to follow the summary and remarks, "
                    f"but got: {tail_element}"
                )

            for field in tail_element.children:
                assert len(field.children) == 2
                field_name, field_body = field.children
                assert isinstance(field_name, docutils.nodes.field_name)
                assert isinstance(field_body, docutils.nodes.field_body)

                # region Generate field body

                body_blocks = []  # type: List[str]
                for body_child in field_body.children:
                    body_block, error = renderer.transform(body_child)
                    if error:
                        return None, error

                    assert body_block is not None

                    body_blocks.append(body_block)

                if len(body_blocks) == 0:
                    body = ""
                elif len(body_blocks) == 1:
                    body = body_blocks[0]
                else:
                    body = "\n".join(
                        f"<para>{body_block}</para>" for body_block in body_blocks
                    )

                # endregion

                # region Generate tags in the description

                assert len(field_name.children) == 1 and isinstance(
                    field_name.children[0], docutils.nodes.Text
                )

                name = field_name.children[0].astext()
                name_parts = name.split()
                if len(name_parts) > 2:
                    # BEFORE-RELEASE (mristin, 2021-12-13): test
                    return (
                        None,
                        f"Expected one or two parts in a field name, "
                        f"but got: {field_name}",
                    )

                if len(name_parts) == 1:
                    directive = name_parts[0]
                    if directive in ("return", "returns"):
                        body_indented = textwrap.indent(body, I)
                        blocks.append(
                            Stripped(f"<returns>\n{body_indented}\n</returns>")
                        )
                    else:
                        return None, f"Unhandled directive: {directive}"

                elif len(name_parts) == 2:
                    directive, directive_arg = name_parts

                    if directive == "param":
                        arg_name = csharp_naming.argument_name(directive_arg)

                        if body != "":
                            indented_body = textwrap.indent(body, I)
                            blocks.append(
                                Stripped(
                                    f"<param name={xml.sax.saxutils.quoteattr(arg_name)}>\n"
                                    f"{indented_body}\n"
                                    f"</param>"
                                )
                            )
                        else:
                            blocks.append(
                                Stripped(
                                    f"<param name={xml.sax.saxutils.quoteattr(arg_name)}>"
                                    f"</param>"
                                )
                            )
                    else:
                        return None, f"Unhandled directive: {directive}"
                else:
                    return (
                        None,
                        f"Expected one or two parts in a field name, "
                        f"but got: {field_name}",
                    )

                # endregion

        # fmt: off
        text = '\n'.join(
            f'/// {line}'
            for line in '\n'.join(blocks).splitlines()
        )
        # fmt: on
        return text, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def generate_comment(
    description: intermediate.Description,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate a documentation comment based on the docstring."""
    if len(description.document.children) == 0:
        return Stripped(""), None

    renderer = _ElementRenderer()
    text, error = renderer.transform(description.document)
    if error:
        return None, Error(description.node, error)

    assert text is not None
    return Stripped(text), None



"""Generate JSON schema corresponding to the meta-model."""
import collections
import json
from typing import (
    TextIO,
    Any,
    MutableMapping,
    Optional,
    Tuple,
    List,
    Sequence,
    Mapping,
    Set,
)

from icontract import ensure

from aas_core_codegen import (
    naming,
    specific_implementations,
    intermediate,
    run,
    infer_for_schema,
)
from aas_core_codegen.common import Stripped, Error, assert_never, Identifier


def _define_for_enumeration(
    enumeration: intermediate.Enumeration,
) -> MutableMapping[str, Any]:
    """
    Generate the definition for an ``enumeration``.

    The list of definitions is to be *extended* with the resulting mapping.
    """
    definition = collections.OrderedDict()  # type: MutableMapping[str, Any]
    definition["type"] = "string"
    definition["enum"] = [literal.value for literal in enumeration.literals]

    model_type = naming.json_model_type(enumeration.name)

    return collections.OrderedDict([(model_type, definition)])


_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "boolean",
    intermediate.PrimitiveType.INT: "integer",
    intermediate.PrimitiveType.FLOAT: "number",
    intermediate.PrimitiveType.STR: "string",
    intermediate.PrimitiveType.BYTEARRAY: "string",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)


def _define_primitive_type(
    primitive_type: intermediate.PrimitiveType,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> MutableMapping[str, Any]:
    """
    Generate the type definition for the ``primitive_type``.

    The associated constraints are included in the definition, if specified.
    """
    type_definition = collections.OrderedDict(
        [("type", _PRIMITIVE_MAP[primitive_type])]
    )  # type: MutableMapping[str, Any]

    if (
        primitive_type
        in (intermediate.PrimitiveType.STR, intermediate.PrimitiveType.BYTEARRAY)
        and len_constraint is not None
    ):
        if len_constraint.min_value is not None:
            type_definition["minLength"] = len_constraint.min_value

        if len_constraint.max_value is not None:
            type_definition["maxLength"] = len_constraint.max_value

    if (
        primitive_type == intermediate.PrimitiveType.STR
        and pattern_constraints is not None
        and len(pattern_constraints) > 0
    ):
        if len(pattern_constraints) == 1:
            type_definition["pattern"] = pattern_constraints[0].pattern
        else:
            all_of = [type_definition]  # type: List[MutableMapping[str, Any]]

            for pattern_constraint in pattern_constraints:
                all_of.append(
                    collections.OrderedDict([("pattern", pattern_constraint.pattern)])
                )

            type_definition = collections.OrderedDict([("allOf", all_of)])

    return type_definition


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_type(
    type_annotation: intermediate.TypeAnnotation,
    ref_association: intermediate.ClassUnion,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[Error]]:
    """
    Generate the type definition for ``type_annotation``.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        type_definition = _define_primitive_type(
            primitive_type=type_annotation.a_type,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )

        return type_definition, None

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        model_type = naming.json_model_type(type_annotation.symbol.name)

        if isinstance(type_annotation.symbol, intermediate.Enumeration):
            return (
                collections.OrderedDict([("$ref", f"#/definitions/{model_type}")]),
                None,
            )

        elif isinstance(type_annotation.symbol, intermediate.ConstrainedPrimitive):
            type_definition = collections.OrderedDict(
                [("$ref", f"#/definitions/{model_type}")]
            )

            if (
                type_annotation.symbol.constrainee
                in (
                    intermediate.PrimitiveType.STR,
                    intermediate.PrimitiveType.BYTEARRAY,
                )
                and len_constraint is not None
            ):
                if len_constraint.min_value is not None:
                    type_definition["minLength"] = len_constraint.min_value

                if len_constraint.max_value is not None:
                    type_definition["maxLength"] = len_constraint.max_value

            if (
                type_annotation.symbol.constrainee is intermediate.PrimitiveType.STR
                and pattern_constraints is not None
                and len(pattern_constraints) > 0
            ):
                if len(pattern_constraints) == 1:
                    type_definition["pattern"] = pattern_constraints[0].pattern
                else:
                    all_of = [type_definition]  # type: List[MutableMapping[str, Any]]

                    for pattern_constraint in pattern_constraints:
                        all_of.append(
                            collections.OrderedDict(
                                [("pattern", pattern_constraint.pattern)]
                            )
                        )

                    type_definition = collections.OrderedDict([("allOf", all_of)])

            return type_definition, None

        elif isinstance(type_annotation.symbol, intermediate.Class):
            if type_annotation.symbol.interface is not None:
                return (
                    collections.OrderedDict(
                        [("$ref", f"#/definitions/{model_type}_abstract")]
                    ),
                    None,
                )
            else:
                return (
                    collections.OrderedDict([("$ref", f"#/definitions/{model_type}")]),
                    None,
                )

        else:
            assert_never(type_annotation.symbol)

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        # NOTE (mristin, 2021-12-02):
        # We do not propagate the inference of constraints to sub-lists
        # so in this case we set all the constraint on the ``items`` to ``None``.
        # This behavior might change in the future when we encounter such
        # constraints and have more context available.

        items_type_definition, items_error = _define_type(
            type_annotation=type_annotation.items,
            ref_association=ref_association,
            len_constraint=None,
            pattern_constraints=None,
        )

        if items_error is not None:
            return None, items_error

        assert items_type_definition is not None

        type_definition = collections.OrderedDict(
            [("type", "array"), ("items", items_type_definition)]
        )

        if len_constraint is not None:
            if len_constraint.min_value is not None:
                assert (
                    "minItems" not in type_definition
                ), "Unexpected property 'minItems' in the JSON type definition"

                type_definition["minItems"] = len_constraint.min_value

            if len_constraint.max_value is not None:
                assert (
                    "maxItems" not in type_definition
                ), "Unexpected property 'maxItems' in the JSON type definition"

                type_definition["maxItems"] = len_constraint.max_value

        return type_definition, None

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        raise NotImplementedError(
            f"(mristin, 2021-11-10):\n"
            f"Nested optional values are unexpected in the JSON schema. "
            f"We did not implement them at the moment since we need more information "
            f"about the context.\n\n"
            f"This feature needs yet to be implemented.\n\n"
            f"{type_annotation=}"
        )

    elif isinstance(type_annotation, intermediate.RefTypeAnnotation):
        model_type = naming.json_model_type(ref_association.name)

        return collections.OrderedDict([("$ref", f"#/definitions/{model_type}")]), None

    else:
        raise NotImplementedError(
            f"(mristin, 2021-11-10):\n"
            f"We implemented only a subset of possible type annotations "
            f"to be represented in a JSON schema since we lacked more information "
            f"about the context.\n\n"
            f"This feature needs yet to be implemented.\n\n"
            f"{type_annotation=}"
        )


def _define_for_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    pattern_verifications_by_name: infer_for_schema.PatternVerificationsByName,
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[List[Error]]]:
    """
    Generate the JSON definitions based on the ``constrained_primitive``.

    The list of definitions is to be *extended* with the resulting mapping.
    """
    all_of = []  # type: List[MutableMapping[str, Any]]

    # region Inheritance

    for inheritance in constrained_primitive.inheritances:
        all_of.append(
            {"$ref": f"#/definitions/{naming.json_model_type(inheritance.name)}"}
        )

    # endregion

    # region Constraints

    errors = []  # type: List[Error]

    (
        len_constraint,
        len_constraint_errors,
    ) = infer_for_schema.infer_len_constraint_of_self(
        constrained_primitive=constrained_primitive
    )

    if len_constraint_errors is not None:
        errors.extend(len_constraint_errors)

    if len(errors) > 0:
        return None, errors

    assert len_constraint is not None

    pattern_constraints = infer_for_schema.infer_patterns_on_self(
        constrained_primitive=constrained_primitive,
        pattern_verifications_by_name=pattern_verifications_by_name,
    )

    type_definition = _define_primitive_type(
        primitive_type=constrained_primitive.constrainee,
        len_constraint=len_constraint,
        pattern_constraints=pattern_constraints,
    )

    all_of.append(type_definition)

    # endregion

    model_type = naming.json_model_type(constrained_primitive.name)

    result = collections.OrderedDict()  # type: MutableMapping[str, Any]
    assert (
        len(all_of) >= 1
    ), "At least the type definition for the primitive type expected"
    result[model_type] = {"allOf": all_of} if len(all_of) > 1 else all_of[0]

    return result, None


# fmt: off
@ensure(
    lambda result:
    (result[0] is not None and result[1] is not None) ^ (result[2] is not None)
)
# fmt: on
def _define_properties_and_required(
    cls: intermediate.Class,
    ref_association: intermediate.ClassUnion,
    pattern_verifications_by_name: infer_for_schema.PatternVerificationsByName,
) -> Tuple[
    Optional[MutableMapping[str, Any]],
    Optional[List[Identifier]],
    Optional[List[Error]],
]:
    """
    Define the ``properties`` and ``required`` part for the given class ``cls``.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    errors = []  # type: List[Error]

    (
        len_constraints_by_property,
        len_constraints_errors,
    ) = infer_for_schema.infer_len_constraints_by_class_properties(cls=cls)

    if len_constraints_errors is not None:
        errors.extend(len_constraints_errors)

    if len(errors) > 0:
        return None, None, errors

    assert len_constraints_by_property is not None

    pattern_constraints_by_property = (
        infer_for_schema.infer_patterns_by_class_properties(
            cls=cls, pattern_verifications_by_name=pattern_verifications_by_name
        )
    )

    properties = collections.OrderedDict()  # type: MutableMapping[str, Any]
    required = []  # type: List[Identifier]

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        prop_name = naming.json_property(prop.name)

        len_constraint = len_constraints_by_property.get(prop, None)
        pattern_constraints = pattern_constraints_by_property.get(prop, None)

        # noinspection PyUnusedLocal
        type_anno = None  # type: Optional[intermediate.TypeAnnotation]
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            type_anno = prop.type_annotation.value
        else:
            type_anno = prop.type_annotation

            required.append(prop_name)

        assert type_anno is not None
        type_definition, error = _define_type(
            type_annotation=type_anno,
            ref_association=ref_association,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )

        if error is not None:
            errors.append(error)
        else:
            assert type_definition is not None
            properties[prop_name] = type_definition

    if len(errors) > 0:
        return None, None, errors

    return properties, required, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_for_class(
    cls: intermediate.Class,
    ref_association: intermediate.ClassUnion,
    ids_of_classes_in_properties: Set[int],
    pattern_verifications_by_name: infer_for_schema.PatternVerificationsByName,
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[List[Error]]]:
    """
    Generate the JSON definitions based on the class ``cls``.

    The list of definitions is to be *extended* with the resulting mapping.

    The ``ref_association`` indicates which symbol to use for representing references
    within an AAS.
    """
    all_of = []  # type: List[MutableMapping[str, Any]]

    # region Inheritance

    for inheritance in cls.inheritances:
        all_of.append(
            {"$ref": f"#/definitions/{naming.json_model_type(inheritance.name)}"}
        )

    # endregion

    # region Properties

    errors = []  # type: List[Error]

    properties, required, properties_error = _define_properties_and_required(
        cls=cls,
        ref_association=ref_association,
        pattern_verifications_by_name=pattern_verifications_by_name,
    )

    if properties_error is not None:
        errors.extend(properties_error)

    if len(errors) > 0:
        return None, errors

    assert properties is not None
    assert required is not None

    if len(properties) > 0:
        definition = collections.OrderedDict(
            [("type", "object"), ("properties", properties)]
        )

        if len(required) > 0:
            definition["required"] = required

        all_of.append(definition)

    # endregion

    model_type = naming.json_model_type(cls.name)

    result = collections.OrderedDict()  # type: MutableMapping[str, Any]
    result[model_type] = {"allOf": all_of} if len(all_of) > 0 else {"type": "object"}

    # region Define the abstract part

    # NOTE (mristin, 2022-01-02):
    # We generate the "*_abstract" definition of a class only if it is used to specify
    # the type of one or more properties in the meta-model. Otherwise, we can ignore
    # the abstract definition as it wouldn't be used during the validation.

    if len(cls.concrete_descendants) > 0 and id(cls) in ids_of_classes_in_properties:
        model_type_abstract = f"{model_type}_abstract"

        any_of = [
            {"$ref": f"#/definitions/{naming.json_model_type(descendant.name)}"}
            for descendant in cls.concrete_descendants
        ]  # type: List[MutableMapping[str, Any]]

        result[model_type_abstract] = {"anyOf": any_of}

    # endregion

    return result, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the JSON schema based on the ``symbol_table."""
    schema_base_key = specific_implementations.ImplementationKey("schema_base.json")

    schema_base_json = spec_impls.get(schema_base_key, None)
    if schema_base_json is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the base schema "
                f"is missing: {schema_base_key}",
            )
        ]

    # noinspection PyUnusedLocal
    schema = None  # type: Optional[MutableMapping[str, Any]]

    try:
        # noinspection PyTypeChecker
        schema = json.loads(schema_base_json, object_pairs_hook=collections.OrderedDict)
    except json.JSONDecodeError as err:
        return None, [
            Error(
                None, f"Failed to parse the base schema from {schema_base_key}: {err}"
            )
        ]

    assert schema is not None

    if "definitions" in schema:
        return None, [
            Error(
                None, "The property ``definitions`` unexpected in the base JSON schema"
            )
        ]

    errors = []  # type: List[Error]

    definitions = collections.OrderedDict()

    pattern_verifications_by_name = infer_for_schema.map_pattern_verifications_by_name(
        verifications=symbol_table.verification_functions
    )

    ids_of_classes_in_properties = intermediate.collect_ids_of_classes_in_properties(
        symbol_table=symbol_table
    )

    for symbol in symbol_table.symbols:
        # Key-value pairs to extend the definitions
        extension = None  # type: Optional[Mapping[str, Any]]

        if isinstance(symbol, intermediate.Class) and symbol.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"{symbol.name}.json"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        symbol.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            try:
                # noinspection PyTypeChecker
                extension = json.loads(code, object_pairs_hook=collections.OrderedDict)
            except Exception as err:
                errors.append(
                    Error(
                        symbol.parsed.node,
                        f"Failed to parse the JSON out of "
                        f"the specific implementation {implementation_key}: {err}",
                    )
                )
                continue

            if not isinstance(extension, dict):
                errors.append(
                    Error(
                        symbol.parsed.node,
                        f"Expected the implementation-specific snippet "
                        f"at {implementation_key} to be a JSON object, "
                        f"but got: {type(extension)}",
                    )
                )
                continue
        else:
            if isinstance(symbol, intermediate.Enumeration):
                extension = _define_for_enumeration(enumeration=symbol)

            elif isinstance(symbol, intermediate.ConstrainedPrimitive):
                extension, definition_errors = _define_for_constrained_primitive(
                    constrained_primitive=symbol,
                    pattern_verifications_by_name=pattern_verifications_by_name,
                )

                if definition_errors is not None:
                    errors.extend(definition_errors)
                    continue

            elif isinstance(symbol, intermediate.Class):
                extension, definition_errors = _define_for_class(
                    cls=symbol,
                    ref_association=symbol_table.ref_association,
                    ids_of_classes_in_properties=ids_of_classes_in_properties,
                    pattern_verifications_by_name=pattern_verifications_by_name,
                )

                if definition_errors is not None:
                    errors.extend(definition_errors)
                    continue

            else:
                assert_never(symbol)

        assert extension is not None
        for identifier, definition in extension.items():
            if identifier in definitions:
                errors.append(
                    Error(
                        symbol.parsed.node,
                        f"One of the JSON definitions, {identifier}, "
                        f"for the symbol {symbol.name} has been "
                        f"already provided in the definitions; "
                        f"did you already define it in another implementation-specific "
                        f"snippet?",
                    )
                )
                continue
            else:
                definitions[identifier] = definition

    if len(errors) > 0:
        return None, errors

    schema["definitions"] = definitions

    return Stripped(json.dumps(schema, indent=2)), None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the JSON Schema "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "schema.json"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the JSON schema to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



import unittest

import aas_core_codegen.csharp.common as csharp_common


class TestStringLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual('""', csharp_common.string_literal(""))

    def test_identifier(self) -> None:
        self.assertEqual('"something"', csharp_common.string_literal("something"))

    def test_single_quote(self) -> None:
        self.assertEqual('"some\'thing"', csharp_common.string_literal("some'thing"))

    def test_double_quotes(self) -> None:
        self.assertEqual('"some\\"thing"', csharp_common.string_literal('some"thing'))

    def test_special_chars(self) -> None:
        self.assertEqual(
            '"some\\n\\\\thing"', csharp_common.string_literal("some\n\\thing")
        )


if __name__ == "__main__":
    unittest.main()



"""Provide types for the references in the docstrings."""

# pylint: disable=keyword-arg-before-vararg

from typing import Union

import docutils
import docutils.nodes
from icontract import require

from aas_core_codegen.intermediate._types import (
    Symbol,
    Property,
    Enumeration,
    EnumerationLiteral,
    ClassUnion,
)


class SymbolReference(
    docutils.nodes.Inline, docutils.nodes.TextElement  # type: ignore
):
    """Represent a reference in the documentation to a symbol in the symbol table."""

    def __init__(  # type: ignore
        self,
        symbol: Symbol,
        rawsource="",
        text="",
        *children,
        **attributes,
    ) -> None:
        """Initialize with the given symbol and propagate the rest to the parent."""
        self.symbol = symbol
        docutils.nodes.TextElement.__init__(
            self, rawsource, text, *children, **attributes
        )


class PropertyReference:
    """Model a reference to a property, usually used in the docstrings."""

    @require(lambda cls, prop: id(prop) in cls.property_id_set)
    def __init__(self, cls: ClassUnion, prop: Property) -> None:
        self.cls = cls
        self.prop = prop


class EnumerationLiteralReference:
    """Model a reference to an enumeration literal, usually used in the docstrings."""

    @require(lambda symbol, literal: id(literal) in symbol.literal_id_set)
    def __init__(self, symbol: Enumeration, literal: EnumerationLiteral) -> None:
        self.symbol = symbol
        self.literal = literal


class AttributeReference(
    docutils.nodes.Inline, docutils.nodes.TextElement  # type: ignore
):
    """
    Represent a reference in the documentation to an "attribute".

    The attribute, in this context, refers to the role ``:attr:``. The references
    implies either a reference to a property of a class or a literal of an enumeration.
    """

    def __init__(  # type: ignore
        self,
        reference: Union[PropertyReference, EnumerationLiteralReference],
        rawsource="",
        text="",
        *children,
        **attributes,
    ) -> None:
        """Initialize with ``property_name`` and propagate the rest to the parent."""
        self.reference = reference
        docutils.nodes.TextElement.__init__(
            self, rawsource, text, *children, **attributes
        )


class ArgumentReference(
    docutils.nodes.Inline, docutils.nodes.TextElement  # type: ignore
):
    """
    Represent a reference in the documentation to a method argument ("parameter").

    The argument, in this context, refers to the role ``:paramref:``.
    """

    def __init__(  # type: ignore
        self,
        reference: str,
        rawsource="",
        text="",
        *children,
        **attributes,
    ) -> None:
        """Initialize with ``reference`` and propagate the rest to the parent."""
        self.reference = reference
        docutils.nodes.TextElement.__init__(
            self, rawsource, text, *children, **attributes
        )



"""Generate names from our ``Pasal_case`` for the respective targets."""
from typing import List

from icontract import ensure

from aas_core_codegen.common import Identifier

UPPERCASE_ABBREVIATION_SET = {"IRI", "IRDI", "IEC", "ID", "URL"}


def json_property(identifier: Identifier) -> Identifier:
    """
    Generate a JSON name of a property based on its meta-model ``identifier``.

    >>> json_property(Identifier("something"))
    'something'

    >>> json_property(Identifier("something_to_URL"))
    'somethingToURL'
    """
    parts = identifier.split("_")

    if len(parts) == 1:
        return Identifier(parts[0].lower())

    cased_parts = [parts[0].lower()]  # type: List[str]
    for part in parts[1:]:
        if part.upper() in UPPERCASE_ABBREVIATION_SET:
            cased_parts.append(part.upper())
        else:
            cased_parts.append(part.capitalize())

    return Identifier("".join(cased_parts))


# fmt: off
@ensure(
    lambda result: "_" not in result
) # This post-condition avoids naming conflicts with prefixing in the JSON schema.
# fmt: on
def json_model_type(identifier: Identifier) -> Identifier:
    """
    Generate the ``modelType`` of the class based on its meta-model ``identifier``.

    >>> json_model_type(Identifier("something"))
    'Something'

    >>> json_model_type(Identifier("Data_type_IEC_61360"))
    'DataTypeIEC61360'
    """
    parts = identifier.split("_")

    cased_parts = []  # type: List[str]
    for part in parts:
        if part.upper() in UPPERCASE_ABBREVIATION_SET:
            cased_parts.append(part.upper())
        else:
            cased_parts.append(part.capitalize())

    return Identifier("".join(cased_parts))


def xml_class_name(identifier: Identifier) -> Identifier:
    """
    Generate the XML tag name for the given class based on its ``identifier``.

    >>> xml_class_name(Identifier("something"))
    'something'

    >>> xml_class_name(Identifier("URL_to_something"))
    'urlToSomething'
    """
    parts = identifier.split("_")
    assert (
        len(parts) >= 1
    ), f"Expected at least one part for the valid identifier: {identifier}"

    if len(parts) == 1:
        return Identifier(parts[0].lower())

    # pylint: disable=consider-using-f-string
    return Identifier(
        "{}{}".format(
            parts[0].lower(), "".join(part.capitalize() for part in parts[1:])
        )
    )


def xml_property(identifier: Identifier) -> Identifier:
    """
    Generate the XML name for the given property based on its ``identifier``.

    >>> xml_property(Identifier("something"))
    'something'

    >>> xml_property(Identifier("URL_to_something"))
    'urlToSomething'
    """
    parts = identifier.split("_")
    assert (
        len(parts) >= 1
    ), f"Expected at least one part for the valid identifier: {identifier}"

    if len(parts) == 1:
        return Identifier(parts[0].lower())

    # pylint: disable=consider-using-f-string
    return Identifier(
        "{}{}".format(
            parts[0].lower(), "".join(part.capitalize() for part in parts[1:])
        )
    )



"""Encapsulate the entry point to different generators."""
import pathlib
import textwrap
from typing import Sequence, TextIO

from icontract import require

from aas_core_codegen import specific_implementations, intermediate
from aas_core_codegen.common import LinenoColumner


class Context:
    """Represent the context of a code generation."""

    @require(lambda model_path: model_path.exists() and model_path.is_file())
    @require(lambda output_dir: output_dir.exists() and output_dir.is_dir())
    def __init__(
        self,
        model_path: pathlib.Path,
        symbol_table: intermediate.SymbolTable,
        spec_impls: specific_implementations.SpecificImplementations,
        lineno_columner: LinenoColumner,
        output_dir: pathlib.Path,
    ) -> None:
        """Initialize with the given values."""
        self.model_path = model_path
        self.symbol_table = symbol_table
        self.spec_impls = spec_impls
        self.lineno_columner = lineno_columner
        self.output_dir = output_dir


@require(
    lambda errors: all(
        len(error) > 0 and not error.startswith("\n")
        # This is necessary so that we do not have double bullet point.
        and not error.startswith("*") and not error.endswith("\n")
        for error in errors
    )
)
@require(lambda message: not message.endswith(":"))
@require(lambda message: not message.endswith("\n"))
@require(lambda message: not message.startswith("\n") and not message.startswith("*"))
# fmt: on
def write_error_report(message: str, errors: Sequence[str], stderr: TextIO) -> None:
    """
    Write the report (main ``message`` and details as ``errors``) to ``stderr``.

    This method helps us to have a unified way of showing errors.
    """
    stderr.write(f"{message}:\n")
    for error in errors:
        indented = textwrap.indent(error, "  ")
        indented = "* " + indented[2:]
        stderr.write(f"{indented}\n")



"""Provide our own abstract syntax tree for contract transpilation."""
import abc
import ast
import enum
from typing import Sequence, Union, Generic, TypeVar, List, Optional

from icontract import DBC

from aas_core_codegen import stringify
from aas_core_codegen.common import Identifier, assert_never

T = TypeVar("T")


class Node(abc.ABC):
    """Represent an abstract node of our syntax tree."""

    def __init__(self, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        self.original_node = original_node

    def __str__(self) -> str:
        """Provide a human-readable representation of the instance."""
        return dump(self)

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()


class Statement(Node):
    """Represent a statement in a program."""

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()


class Expression(Node):
    """Represent an expression in our abstract syntax tree."""

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()


class Member(Expression):
    """
    Represent a member of an instance.

    A member is either a property or a method.
    """

    def __init__(
        self, instance: "Expression", name: Identifier, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.instance = instance
        self.name = name

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_member(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_member(self)


class Comparator(enum.Enum):
    """List comparison operands."""

    LT = "LT"
    LE = "LE"
    GT = "GT"
    GE = "GE"
    EQ = "EQ"
    NE = "NE"


class Comparison(Expression):
    """Represent a comparison operation."""

    def __init__(
        self,
        left: "Expression",
        op: Comparator,
        right: "Expression",
        original_node: ast.AST,
    ) -> None:
        Expression.__init__(self, original_node=original_node)
        self.left = left
        self.op = op
        self.right = right

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_comparison(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_comparison(self)


class Implication(Expression):
    """Represent an implication of the form ``A => B``."""

    def __init__(
        self, antecedent: "Expression", consequent: "Expression", original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.antecedent = antecedent
        self.consequent = consequent

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_implication(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_implication(self)


class MethodCall(Expression):
    """Represent a method call."""

    def __init__(
        self, member: Member, args: Sequence["Expression"], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.member = member
        self.args = args

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_method_call(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_method_call(self)


class Name(Expression):
    """Represent an access to a variable with the given name."""

    def __init__(self, identifier: Identifier, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.identifier = identifier

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_name(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_name(self)


class FunctionCall(Expression):
    """Represent a function call."""

    def __init__(
        self, name: Name, args: Sequence["Expression"], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.name = name
        self.args = args

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_function_call(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_function_call(self)


class Constant(Expression):
    """Represent a constant value."""

    def __init__(
        self, value: Union[bool, int, float, str], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_constant(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_constant(self)


class IsNone(Expression):
    """Represent a check whether something ``is None``."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_is_none(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_is_none(self)


class IsNotNone(Expression):
    """Represent a check whether something ``is not None``."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_is_not_none(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_is_not_none(self)


class And(Expression):
    """Represent a conjunction."""

    def __init__(self, values: Sequence[Expression], original_node: ast.AST) -> None:
        Expression.__init__(self, original_node=original_node)
        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_and(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_and(self)


class Or(Expression):
    """Represent a disjunction."""

    def __init__(self, values: Sequence[Expression], original_node: ast.AST) -> None:
        Expression.__init__(self, original_node=original_node)
        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_or(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_or(self)


class Declaration(Statement):
    """Represent a variable declaration with a walrus operator, ``:=``."""

    def __init__(
        self, identifier: Identifier, value: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, original_node=original_node)
        self.identifier = identifier
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_declaration(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_declaration(self)


class ExpressionWithDeclarations(Expression):
    """
    Represent a declaration of a local variable followed by the expression.

    This abstract the code patterns such as ``(x := ..., len(x) > 0)[1]``, similar to,
    *e.g.*, short ``if`` statements in Golang.
    """

    def __init__(
        self,
        declarations: Sequence[Declaration],
        expression: Expression,
        original_node: ast.AST,
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.declarations = declarations
        self.expression = expression

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_expression_with_declarations(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_expression_with_declarations(self)


class FormattedValue(Node):
    """Represent a formatted value in a :py:class`JoinedStr`."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Node.__init__(self, original_node=original_node)

        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_formatted_value(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_formatted_value(self)


class JoinedStr(Expression):
    """Represent a string interpolation."""

    def __init__(
        self, values: Sequence[Union[str, FormattedValue]], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)

        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_joined_str(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_joined_str(self)


class Assignment(Statement):
    """Represent an assignment of a single value to a single target."""

    def __init__(
        self, target: Expression, value: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, original_node=original_node)
        self.target = target
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_assignment(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_assignment(self)


class Return(Statement):
    """Represent a return statement with a single return value."""

    def __init__(self, value: Optional[Expression], original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_return(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_return(self)


class Visitor(DBC):
    """
    Visit all the nodes in the AST.

    The default action for the generic visitor is to descend recursively into all the
    children nodes.
    """

    def visit(self, node: Node) -> None:
        """Dispatch to the appropriate visitation method."""
        node.visit(self)

    def visit_member(self, node: Member) -> None:
        """Visit a member."""
        self.visit(node.instance)

    def visit_comparison(self, node: Comparison) -> None:
        """Visit a comparison."""
        self.visit(node.left)
        self.visit(node.right)

    def visit_implication(self, node: Implication) -> None:
        """Visit an implication."""
        self.visit(node.antecedent)
        self.visit(node.consequent)

    def visit_method_call(self, node: MethodCall) -> None:
        """Visit a method call."""
        self.visit(node.member)
        for arg in node.args:
            self.visit(arg)

    def visit_function_call(self, node: FunctionCall) -> None:
        """Visit a function call."""
        for arg in node.args:
            self.visit(arg)

    def visit_constant(self, node: Constant) -> None:
        """Visit a constant."""
        pass

    def visit_is_none(self, node: IsNone) -> None:
        """Visit an is-none check."""
        self.visit(node.value)

    def visit_is_not_none(self, node: IsNotNone) -> None:
        """Visit an is-not-none check."""
        self.visit(node.value)

    def visit_name(self, node: Name) -> None:
        """Visit a variable access."""
        pass

    def visit_and(self, node: And) -> None:
        """Visit a conjunction."""
        for value in node.values:
            self.visit(value)

    def visit_or(self, node: Or) -> None:
        """Visit a disjunction."""
        for value in node.values:
            self.visit(value)

    def visit_declaration(self, node: Declaration) -> None:
        """Visit a variable declaration."""
        self.visit(node.value)

    def visit_expression_with_declarations(
        self, node: ExpressionWithDeclarations
    ) -> None:
        """Visit an expression with variable declarations."""
        for declaration in node.declarations:
            self.visit(declaration)

        self.visit(node.expression)

    def visit_formatted_value(self, node: FormattedValue) -> None:
        """Visit a formatted value in a joined string."""
        self.visit(node.value)

    def visit_joined_str(self, node: JoinedStr) -> None:
        """Visit a string interpolation."""
        for value in node.values:
            if isinstance(value, str):
                pass
            elif isinstance(value, FormattedValue):
                self.visit(value)
            else:
                assert_never(value)

    def visit_assignment(self, node: Assignment) -> None:
        """Visit an assignment statement."""
        self.visit(node.target)
        self.visit(node.value)

    def visit_return(self, node: Return) -> None:
        """Visit a return statement."""
        if node.value is not None:
            self.visit(node.value)


class Transformer(Generic[T], DBC):
    """Transform our AST into something."""

    def transform(self, node: Node) -> T:
        """Dispatch to the appropriate transformation method."""
        return node.transform(self)

    @abc.abstractmethod
    def transform_member(self, node: Member) -> T:
        """Transform a member to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_comparison(self, node: Comparison) -> T:
        """Transform a comparison to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_implication(self, node: Implication) -> T:
        """Transform an implication to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_method_call(self, node: MethodCall) -> T:
        """Transform a method call into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_function_call(self, node: FunctionCall) -> T:
        """Transform a function call into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_constant(self, node: Constant) -> T:
        """Transform a constant into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_is_none(self, node: IsNone) -> T:
        """Transform an is-none check into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_is_not_none(self, node: IsNotNone) -> T:
        """Transform an is-not-none check into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_name(self, node: Name) -> T:
        """Transform a variable access into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_and(self, node: And) -> T:
        """Transform a conjunction into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_or(self, node: Or) -> T:
        """Transform a disjunction into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_declaration(self, node: Declaration) -> T:
        """Transform a variable declaration into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_expression_with_declarations(
        self, node: ExpressionWithDeclarations
    ) -> T:
        """Transform an expression with variable declarations into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_formatted_value(self, node: FormattedValue) -> T:
        """Transform a formatted value of a joined string into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_joined_str(self, node: JoinedStr) -> T:
        """Transform a string interpolation into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_assignment(self, node: Assignment) -> T:
        """Transform an assignment into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_return(self, node: Return) -> T:
        """Transform a return statement into something."""
        raise NotImplementedError(f"{node=}")


class RestrictedTransformer(Transformer[T], DBC):
    """
    Transform our AST into something where only a part of the tree is handled.

    This class is helpful for cases where you are sure that certain nodes *can not*
    appear.
    """

    def transform_member(self, node: Member) -> T:
        """Transform a member to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_comparison(self, node: Comparison) -> T:
        """Transform a comparison to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_implication(self, node: Implication) -> T:
        """Transform an implication to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_method_call(self, node: MethodCall) -> T:
        """Transform a method call into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_function_call(self, node: FunctionCall) -> T:
        """Transform a function call into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_constant(self, node: Constant) -> T:
        """Transform a constant into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_is_none(self, node: IsNone) -> T:
        """Transform an is-none check into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_is_not_none(self, node: IsNotNone) -> T:
        """Transform an is-not-none check into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_name(self, node: Name) -> T:
        """Transform a variable access into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_and(self, node: And) -> T:
        """Transform a conjunction into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_or(self, node: Or) -> T:
        """Transform a disjunction into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_declaration(self, node: Declaration) -> T:
        """Transform a variable declaration into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_expression_with_declarations(
        self, node: ExpressionWithDeclarations
    ) -> T:
        """Transform an expression with variable declarations into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_formatted_value(self, node: FormattedValue) -> T:
        """Transform a formatted value in a joined string into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_joined_str(self, node: JoinedStr) -> T:
        """Transform a string interpolation into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_assignment(self, node: Assignment) -> T:
        """Transform an assignment into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_return(self, node: Return) -> T:
        """Transform a return statement into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")


class _StringifyTransformer(Transformer[stringify.Entity]):
    """Transform a node into a stringifiable representation."""

    def transform(self, node: Node) -> stringify.Entity:
        """Dispatch to the appropriate transformation method."""
        result = node.transform(self)
        stringify.assert_compares_against_dict(entity=result, obj=node)
        return node.transform(self)

    def transform_member(self, node: Member) -> stringify.Entity:
        return stringify.Entity(
            name=Member.__name__,
            properties=[
                stringify.Property("instance", self.transform(node.instance)),
                stringify.Property("name", node.name),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_comparison(self, node: Comparison) -> stringify.Entity:
        return stringify.Entity(
            name=Comparison.__name__,
            properties=[
                stringify.Property("left", self.transform(node.left)),
                stringify.Property("op", str(node.op.value)),
                stringify.Property("right", self.transform(node.right)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_implication(self, node: Implication) -> stringify.Entity:
        return stringify.Entity(
            name=Implication.__name__,
            properties=[
                stringify.Property("antecedent", self.transform(node.antecedent)),
                stringify.Property("consequent", self.transform(node.consequent)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_method_call(self, node: MethodCall) -> stringify.Entity:
        return stringify.Entity(
            name=MethodCall.__name__,
            properties=[
                stringify.Property("member", self.transform(node.member)),
                stringify.Property("args", [self.transform(arg) for arg in node.args]),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_function_call(self, node: FunctionCall) -> stringify.Entity:
        return stringify.Entity(
            name=FunctionCall.__name__,
            properties=[
                stringify.Property("name", node.name.identifier),
                stringify.Property("args", [self.transform(arg) for arg in node.args]),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_constant(self, node: Constant) -> stringify.Entity:
        return stringify.Entity(
            name=Constant.__name__,
            properties=[
                stringify.Property("value", node.value),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_is_none(self, node: IsNone) -> stringify.Entity:
        return stringify.Entity(
            name=IsNone.__name__,
            properties=[stringify.Property("value", self.transform(node.value))],
        )

    def transform_is_not_none(self, node: IsNotNone) -> stringify.Entity:
        return stringify.Entity(
            name=IsNotNone.__name__,
            properties=[
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_name(self, node: Name) -> stringify.Entity:
        return stringify.Entity(
            name=Name.__name__,
            properties=[
                stringify.Property("identifier", node.identifier),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_and(self, node: And) -> stringify.Entity:
        return stringify.Entity(
            name=And.__name__,
            properties=[
                stringify.Property(
                    "values", [self.transform(value) for value in node.values]
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_or(self, node: Or) -> stringify.Entity:
        return stringify.Entity(
            name=Or.__name__,
            properties=[
                stringify.Property(
                    "values", [self.transform(value) for value in node.values]
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_declaration(self, node: Declaration) -> stringify.Entity:
        return stringify.Entity(
            name=Declaration.__name__,
            properties=[
                stringify.Property("identifier", node.identifier),
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_expression_with_declarations(
        self, node: ExpressionWithDeclarations
    ) -> stringify.Entity:
        return stringify.Entity(
            name=Declaration.__name__,
            properties=[
                stringify.Property(
                    "declarations",
                    [self.transform(declaration) for declaration in node.declarations],
                ),
                stringify.Property("expression", self.transform(node.expression)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_formatted_value(self, node: FormattedValue) -> stringify.Entity:
        return stringify.Entity(
            name=FormattedValue.__name__,
            properties=[
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_joined_str(self, node: JoinedStr) -> stringify.Entity:
        values = []  # type: List[Union[stringify.Entity, str]]
        for value in node.values:
            if isinstance(value, str):
                values.append(value)
            elif isinstance(value, FormattedValue):
                values.append(self.transform(value))
            else:
                assert_never(value)

        return stringify.Entity(
            name=JoinedStr.__name__,
            properties=[
                stringify.Property(
                    "values",
                    values,
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_assignment(self, node: Assignment) -> stringify.Entity:
        return stringify.Entity(
            name=Assignment.__name__,
            properties=[
                stringify.Property("target", self.transform(node.target)),
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_return(self, node: Return) -> stringify.Entity:
        return stringify.Entity(
            name=Return.__name__,
            properties=[
                stringify.Property(
                    "value",
                    self.transform(node.value) if node.value is not None else None,
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )


def dump(node: Node) -> str:
    """Produce a string representation of the tree."""
    transformer = _StringifyTransformer()
    return stringify.dump(transformer.transform(node=node))



import contextlib
import io
import os
import pathlib
import tempfile
import unittest

import aas_core_codegen.main


class Test_against_recorded(unittest.TestCase):
    # Set this variable to True if you want to re-record the test data,
    # without any checks
    RERECORD = True  # TODO: undo

    def test_cases(self) -> None:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

        parent_case_dir = repo_dir / "test_data" / "jsonschema" / "test_main"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        for case_dir in parent_case_dir.iterdir():
            assert case_dir.is_dir(), case_dir

            model_pth = case_dir / "input/meta_model.py"
            assert model_pth.exists() and model_pth.is_file(), model_pth

            snippets_dir = case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            expected_output_dir = case_dir / "expected_output"

            with contextlib.ExitStack() as exit_stack:
                if Test_against_recorded.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    tmp_dir = tempfile.TemporaryDirectory()
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = aas_core_codegen.main.Parameters(
                    model_path=model_pth,
                    target=aas_core_codegen.main.Target.JSONSCHEMA,
                    snippets_dir=snippets_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()

                return_code = aas_core_codegen.main.execute(
                    params=params, stdout=stdout, stderr=stderr
                )

                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                stdout_pth = expected_output_dir / "stdout.txt"
                normalized_stdout = stdout.getvalue().replace(
                    str(output_dir), "<output dir>"
                )

                if Test_against_recorded.RERECORD:
                    stdout_pth.write_text(normalized_stdout)
                else:
                    self.assertEqual(
                        normalized_stdout, stdout_pth.read_text(), stdout_pth
                    )

                # BEFORE-RELEASE (mristin, 2021-12-13):
                #  check the remainder of the generated files
                for relevant_rel_pth in [
                    pathlib.Path("schema.json"),
                ]:
                    expected_pth = expected_output_dir / relevant_rel_pth
                    output_pth = output_dir / relevant_rel_pth

                    if not output_pth.exists():
                        raise FileNotFoundError(
                            f"The output file is missing: {output_pth}"
                        )

                    if Test_against_recorded.RERECORD:
                        expected_pth.write_text(output_pth.read_text())
                    else:
                        self.assertEqual(
                            expected_pth.read_text(),
                            output_pth.read_text(),
                            f"The files {expected_pth} and {output_pth} do not match.",
                        )


if __name__ == "__main__":
    unittest.main()



import textwrap
import unittest
from typing import MutableMapping, List

import tests.common
from aas_core_codegen import intermediate, infer_for_schema
from aas_core_codegen.common import Identifier


def infer_patterns_on_self_of_class_something(
    source: str,
) -> List[infer_for_schema.PatternConstraint]:
    """Translate the ``source`` into inferred constraints of the class ``Something``."""
    symbol_table, error = tests.common.translate_source_to_intermediate(source=source)
    assert error is None, tests.common.most_underlying_messages(error)
    assert symbol_table is not None
    symbol = symbol_table.must_find(Identifier("Something"))
    assert isinstance(symbol, intermediate.ConstrainedPrimitive)

    pattern_verifications_by_name = infer_for_schema.map_pattern_verifications_by_name(
        verifications=symbol_table.verification_functions
    )

    return infer_for_schema.infer_patterns_on_self(
        constrained_primitive=symbol,
        pattern_verifications_by_name=pattern_verifications_by_name,
    )


class Test_expected(unittest.TestCase):
    def test_no_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        patterns = infer_patterns_on_self_of_class_something(source=source)

        text = infer_for_schema.dump_patterns(patterns)
        self.assertEqual("[]", text)

    def test_single_pattern(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                prefix = "something"
                return match(f"{prefix}-[a-zA-Z]+", text) is not None


            @invariant(lambda self: is_something(self))
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        patterns = infer_patterns_on_self_of_class_something(source=source)

        text = infer_for_schema.dump_patterns(patterns)
        self.assertEqual(
            textwrap.dedent(
                """\
                [
                  PatternConstraint(
                    pattern='something-[a-zA-Z]+')
                ]"""
            ),
            text,
        )

    def test_two_patterns(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None


            @invariant(lambda self: is_acme(self))
            @invariant(lambda self: is_something(self))
            class Something(str):
                pass


            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        patterns = infer_patterns_on_self_of_class_something(source=source)

        text = infer_for_schema.dump_patterns(patterns)
        self.assertEqual(
            textwrap.dedent(
                """\
                [
                  PatternConstraint(
                    pattern='something-[a-zA-Z]+'),
                  PatternConstraint(
                    pattern='.*acme.*')
                ]"""
            ),
            text,
        )

    def test_inheritance(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def is_something(text: str) -> bool:
                return match("something-[a-zA-Z]+", text) is not None

            @verification
            def is_acme(text: str) -> bool:
                return match(".*acme.*", text) is not None

            @invariant(lambda self: is_something(self))
            class Parent(str):
                pass

            @invariant(lambda self: is_acme(self))
            class Something(Parent):
                pass

            class Reference:
                pass

            __book_url__ = "dummy"
            __book_version__ = "dummy"
            associate_ref_with(Reference)
            """
        )

        patterns = infer_patterns_on_self_of_class_something(source=source)

        text = infer_for_schema.dump_patterns(patterns)

        # NOTE (mristin, 2022-01-02):
        # We infer only the constraints as specified in the class itself, and
        # ignore the constraints of the ancestors in *this particular kind of
        # inference*.

        self.assertEqual(
            textwrap.dedent(
                """\
                [
                  PatternConstraint(
                    pattern='.*acme.*')
                ]"""
            ),
            text,
        )


if __name__ == "__main__":
    unittest.main()



"""Provide the intermediate representation of the meta-model."""

from aas_core_codegen.intermediate import _types, _translate, _stringify

TypeAnnotation = _types.TypeAnnotation
TypeAnnotationUnion = _types.TypeAnnotationUnion
PrimitiveType = _types.PrimitiveType
PrimitiveTypeAnnotation = _types.PrimitiveTypeAnnotation
OurTypeAnnotation = _types.OurTypeAnnotation
ListTypeAnnotation = _types.ListTypeAnnotation
OptionalTypeAnnotation = _types.OptionalTypeAnnotation
RefTypeAnnotation = _types.RefTypeAnnotation
Description = _types.Description
Property = _types.Property
Default = _types.Default
DefaultConstant = _types.DefaultConstant
DefaultEnumerationLiteral = _types.DefaultEnumerationLiteral
Argument = _types.Argument
Symbol = _types.Symbol
Invariant = _types.Invariant
Contract = _types.Contract
Snapshot = _types.Snapshot
Contracts = _types.Contracts
Method = _types.Method
ImplementationSpecificMethod = _types.ImplementationSpecificMethod
Constructor = _types.Constructor
Serialization = _types.Serialization
EnumerationLiteral = _types.EnumerationLiteral
Enumeration = _types.Enumeration
ConstrainedPrimitive = _types.ConstrainedPrimitive
Class = _types.Class
ClassUnion = _types.ClassUnion
ConcreteClass = _types.ConcreteClass
AbstractClass = _types.AbstractClass
Verification = _types.Verification
ImplementationSpecificVerification = _types.ImplementationSpecificVerification
PatternVerification = _types.PatternVerification
Signature = _types.Signature
Interface = _types.Interface
SymbolTable = _types.SymbolTable

map_descendability = _types.map_descendability
make_union_of_constructor_arguments = _types.make_union_of_constructor_arguments
collect_ids_of_classes_in_properties = _types.collect_ids_of_classes_in_properties

translate = _translate.translate
errors_if_contracts_for_functions_or_methods_defined = (
    _translate.errors_if_contracts_for_functions_or_methods_defined
)
errors_if_non_implementation_specific_methods = (
    _translate.errors_if_non_implementation_specific_methods
)

dump = _stringify.dump



import textwrap
import unittest.mock

# noinspection PyProtectedMember
import aas_core_codegen.csharp.description as csharp_description
import tests.common
from aas_core_codegen.common import Stripped, Identifier


class TestDescription(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """
        Generate the C# description comment based on ``source``.

        The ``source`` is expected to contain only a single class, a class
        ``Some_class``.
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, f"{error=}"
        assert symbol_table is not None

        some_class = symbol_table.must_find(Identifier("Some_class"))
        assert some_class.description is not None

        code, error = csharp_description.generate_comment(some_class.description)
        assert error is None, f"{error=}"

        assert code is not None

        return code

    def test_empty(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """"""  # Intentionally left empty
                
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual("", comment_code)

    def test_only_summary(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """Do & drink something."""
                
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_class_reference(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """Do & drink :class:`.Some_class`."""
                    
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="SomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_interface_reference(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                @abstract
                class Some_class:
                    """Do & drink :class:`.Some_class`."""
                    
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="ISomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_enumeration_reference(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class(Enum):
                    """Do & drink :class:`.Some_class`."""
                    
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="SomeClass" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_and_remarks(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """
                    Do & drink something.

                    First & remark.

                    Second & remark.
                    """
                    
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>
                /// <remarks>
                /// <para>First &amp; remark.</para>
                /// <para>Second &amp; remark.</para>
                /// </remarks>"""
            ),
            comment_code,
        )

    def test_summary_remarks_and_fields(self) -> None:
        comment_code = TestDescription.render(
            textwrap.dedent(
                '''\
                class Some_class:
                    """
                    Do & drink something.

                    First & remark.

                    :param without_description:
                    :param something: argument description same-line
                    :param another:
                        argument description as paragraph &
                        longer

                        text
                    :returns: some result
                    """
                    
                class Reference:
                    pass
                    
                __book_url__ = "dummy"
                __book_version__ = "dummy"
                associate_ref_with(Reference)
                '''
            )
        )
        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>
                /// <remarks>
                /// First &amp; remark.
                /// </remarks>
                /// <param name="withoutDescription"></param>
                /// <param name="something">
                ///     argument description same-line
                /// </param>
                /// <param name="another">
                ///     <para>argument description as paragraph &amp;
                ///     longer</para>
                ///     <para>text</para>
                /// </param>
                /// <returns>
                ///     some result
                /// </returns>"""
            ),
            comment_code,
        )


if __name__ == "__main__":
    unittest.main()
